// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Neil Jagdish Patel <neil.patel@canonical.com>
 */

#include "config.h"

#include <Nux/Nux.h>
#include <Nux/BaseWindow.h>
#include <Nux/HLayout.h>
#include <Nux/Layout.h>
#include <Nux/WindowCompositor.h>

#include <NuxImage/CairoGraphics.h>
#include <NuxImage/ImageSurface.h>

#include <NuxGraphics/GLThread.h>
#include <NuxGraphics/RenderingPipe.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gio/gdesktopappinfo.h>
#include "ubus-server.h"
#include "UBusMessages.h"

#include "PlacesHomeView.h"

#include "JSONParser.h"
#include "PlacesSimpleTile.h"
#include "PlacesStyle.h"
#include <UnityCore/GLibWrapper.h>
#include <UnityCore/Variant.h>

#include <sstream>
#include <string>
#include <vector>

#include <boost/algorithm/string.hpp>

#define DELTA_DOUBLE_REQUEST 500000000

namespace
{
  const std::string custom_shortcuts_file = "/unity/HomeShortcutsCustomized.json";
}

namespace unity
{

enum
{
  TYPE_PLACE = 0,
  TYPE_EXEC
};

class Shortcut : public PlacesSimpleTile
{
public:
  Shortcut(const char* icon, const char* name, int size)
    : PlacesSimpleTile(icon, name, size),
      _id(0),
      _place_id(NULL),
      _place_section(0),
      _exec(NULL)
  {
    SetDndEnabled(false, false);
  }

  ~Shortcut()
  {
    g_free(_place_id);
    g_free(_exec);
  }

  int      _id;
  gchar*   _place_id;
  guint32  _place_section;
  char*    _exec;
};

PlacesHomeView::PlacesHomeView()
  : _ubus_handle(0)
{
  PlacesStyle* style = PlacesStyle::GetDefault();

  SetName(_("Shortcuts"));
  SetIcon(PKGDATADIR"/shortcuts_group_icon.png");
  SetDrawSeparator(false);

  _layout = new nux::GridHLayout(NUX_TRACKER_LOCATION);
  _layout->SetReconfigureParentLayoutOnGeometryChange(true);
  SetChildLayout(_layout);

  _layout->ForceChildrenSize(true);
  _layout->SetChildrenSize(style->GetHomeTileWidth(), style->GetHomeTileHeight());
  _layout->EnablePartialVisibility(false);
  _layout->SetHeightMatchContent(true);
  _layout->SetHorizontalExternalMargin(32);
  _layout->SetVerticalInternalMargin(32);
  _layout->SetHorizontalInternalMargin(32);
  _layout->SetMinMaxSize((style->GetHomeTileWidth() * 4) + (32 * 5),
                         (style->GetHomeTileHeight() * 2) + 32);

  _ubus_handle = ubus_server_register_interest(ubus_server_get_default(),
                                               UBUS_PLACE_VIEW_SHOWN,
                                               (UBusCallback) &PlacesHomeView::DashVisible,
                                               this);

  //In case the GConf key is invalid (e.g. when an app was uninstalled), we
  //rely on a fallback "whitelist" mechanism instead of showing nothing at all
  _browser_alternatives.push_back("firefox");
  _browser_alternatives.push_back("chromium-browser");
  _browser_alternatives.push_back("epiphany-browser");
  _browser_alternatives.push_back("midori");

  _photo_alternatives.push_back("shotwell");
  _photo_alternatives.push_back("f-spot");
  _photo_alternatives.push_back("gthumb");
  _photo_alternatives.push_back("gwenview");
  _photo_alternatives.push_back("eog");

  _email_alternatives.push_back("evolution");
  _email_alternatives.push_back("thunderbird");
  _email_alternatives.push_back("claws-mail");
  _email_alternatives.push_back("kmail");

  _music_alternatives.push_back("banshee-1");
  _music_alternatives.push_back("rhythmbox");
  _music_alternatives.push_back("totem");
  _music_alternatives.push_back("vlc");

  expanded.connect(sigc::mem_fun(this, &PlacesHomeView::Refresh));

  // Wait for the asynchronous loading of the lenses to complete
  // before refreshing the view for the first time.
  _lenses.lenses_loaded.connect(sigc::bind(sigc::mem_fun(this, &PlacesHomeView::Refresh), (PlacesGroup*) NULL));
}

PlacesHomeView::~PlacesHomeView()
{
  if (_ubus_handle != 0)
    ubus_server_unregister_interest(ubus_server_get_default(), _ubus_handle);
}

void
PlacesHomeView::DashVisible(GVariant* data, void* val)
{
  PlacesHomeView* self = (PlacesHomeView*)val;
  self->Refresh();
}

void
PlacesHomeView::Refresh(PlacesGroup*foo)
{
  PlacesStyle* style = PlacesStyle::GetDefault();
  Shortcut*   shortcut = NULL;
  gchar*      markup = NULL;
  const char* temp = "<big>%s</big>";
  int         icon_size = style->GetHomeTileIconSize();

  _layout->Clear();

  // The following code path allows one to replace the default dash home screen
  // with a custom home screen, the contents of which are described in a JSON
  // file. The file is looked for in the following locations in decreasing order
  // of priority:
  //   - $XDG_CONFIG_HOME/unity/ (defaults to $HOME/.config/unity/)
  //   - $DIR/unity/ for $DIR in $XDG_CONFIG_DIRS
  // The syntax of the file is as follows:
  // ==========================================================================
  // {
  //   "shortcut1": {
  //     "source": $source,
  //     "name": $name,
  //     "name[fr]": $name_in_french,
  //     […]
  //     "icon": $icon
  //   },
  //   "shortcut2": {
  //     […]
  //   },
  //   […]
  // }
  // ==========================================================================
  // The source attribute may either be a desktop file (full path or just its
  // basename if it is located in a standard directory), or a lens file
  // (basename only). This attribute is mandatory.
  // The 'name' attribute is optional. If present, it will override the default
  // display name as advertised by the desktop file or by the lens. The name can
  // be localized in several languages using the square brackets suffix
  // notation, in which case the locale matching the system’s will be used,
  // defaulting to the untranslated 'name' attribute if necessary.
  // The 'icon' attribute is optional. If present, it will override the default
  // icon as advertised by the desktop file or by the lens. It should be a full
  // path name.
  // If the source is a lens, the optional 'filter' attribute allows specifying
  // a filter in the form "$name:$value", e.g. "type:videos" for the files lens.
  // If the source is a lens, the optional 'section' attribute allows specifying
  // a section number (an integer starting at index 0).

  std::string customShortcuts;
  std::ostringstream file;
  // Look for a custom shortcuts file in the user’s home directory first.
  file << g_get_user_config_dir() << custom_shortcuts_file;
  if (g_file_test(file.str().c_str(), G_FILE_TEST_EXISTS))
  {
    customShortcuts = file.str();
  }
  else
  {
    // Fall back on the standard XDG directories.
    gchar** config = (gchar**) g_get_system_config_dirs();
    for (gint i = 0; config[i]; ++i)
    {
      std::ostringstream file;
      file << config[i] << custom_shortcuts_file;
      if (g_file_test(file.str().c_str(), G_FILE_TEST_EXISTS))
      {
        customShortcuts = file.str();
        break;
      }
    }
  }
  json::Parser parser;
  if ((customShortcuts == "") || !parser.Open(customShortcuts))
  {
    // Media Apps
    markup = g_strdup_printf(temp, _("Media Apps"));
    shortcut = new Shortcut(PKGDATADIR"/find_media_apps.png",
                            markup,
                            icon_size);
    shortcut->_id = TYPE_PLACE;
    shortcut->_place_id = g_strdup("applications.lens?filter_type=media");
    shortcut->_place_section = 9;
    _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
    shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
    g_free(markup);

    // Internet Apps
    markup = g_strdup_printf(temp, _("Internet Apps"));
    shortcut = new Shortcut(PKGDATADIR"/find_internet_apps.png",
                            markup,
                            icon_size);
    shortcut->_id = TYPE_PLACE;
    shortcut->_place_id = g_strdup("applications.lens?filter_type=internet");
    shortcut->_place_section = 8;
    _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
    shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
    g_free(markup);

    // More Apps
    markup = g_strdup_printf(temp, _("More Apps"));
    shortcut = new Shortcut(PKGDATADIR"/find_more_apps.png",
                            markup,
                            icon_size);
    shortcut->_id = TYPE_PLACE;
    shortcut->_place_id = g_strdup("applications.lens");
    shortcut->_place_section = 0;
    _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
    shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
    g_free(markup);

    // Find Files
    markup = g_strdup_printf(temp, _("Find Files"));
    shortcut = new Shortcut(PKGDATADIR"/find_files.png",
                            markup,
                            icon_size);
    shortcut->_id = TYPE_PLACE;
    shortcut->_place_id = g_strdup("files.lens");
    shortcut->_place_section = 0;
    _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
    shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
    g_free(markup);

    // Browser
    CreateShortcutFromMime("x-scheme-handler/http", _("Browse the Web"), _browser_alternatives);

    // Photos
    // FIXME: Need to figure out the default
    CreateShortcutFromExec("shotwell", _("View Photos"), _photo_alternatives);

    CreateShortcutFromMime("x-scheme-handler/mailto", _("Check Email"), _email_alternatives);

    CreateShortcutFromMime("audio/x-vorbis+ogg", _("Listen to Music"), _music_alternatives);
  }
  else
  {
    // The custom shortcuts file can contain up to 8 shortcuts.
    for (int i = 0; i < 8; ++i)
    {
      std::ostringstream oss;
      oss << "shortcut" << (i + 1);
      std::string slot = oss.str();
      std::string source;
      parser.ReadString(slot, "source", source);
      if (source != "")
      {
        std::string icon;
        parser.ReadString(slot, "icon", icon);

        std::string name;
        gchar** languages = (gchar**) g_get_language_names();
        for (gint i = 0; languages[i]; ++i)
        {
          std::ostringstream key;
          key << "name[" << languages[i] << "]";
          parser.ReadString(slot, key.str(), name);
          if (name != "")
            break;
        }
        if (name == "")
        {
          parser.ReadString(slot, "name", name);
        }

        if (boost::iends_with(source, ".desktop"))
        {
          GDesktopAppInfo* info = NULL;
          if (boost::istarts_with(source, "/"))
            info = g_desktop_app_info_new_from_filename(source.c_str());
          else
            info = g_desktop_app_info_new(source.c_str());
          if (G_IS_DESKTOP_APP_INFO(info))
          {
            if (name == "")
              name = g_app_info_get_display_name(G_APP_INFO(info));
            if (icon == "")
            {
              gchar* cicon = g_icon_to_string(g_app_info_get_icon(G_APP_INFO(info)));
              icon = cicon;
              g_free(cicon);
            }
            gchar* exec = g_strdup(g_app_info_get_executable(G_APP_INFO(info)));
            markup = g_strdup_printf(temp, name.c_str());
            shortcut = new Shortcut(icon.c_str(), markup, style->GetHomeTileIconSize());
            shortcut->_id = TYPE_EXEC;
            shortcut->_exec = exec;
            _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
            shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
            g_free(markup);
            g_object_unref(info);
          }
        }
        else if (boost::iends_with(source, ".lens"))
        {
          unity::dash::Lens::Ptr lens = _lenses.GetLens(source);
          if (lens)
          {
            if (name == "")
              name = lens->name;
            if (icon == "")
              icon = lens->icon_hint;
          }
          std::string filter;
          parser.ReadString(slot, "filter", filter);
          int section = -1;
          parser.ReadInt(slot, "section", section);
          markup = g_strdup_printf(temp, name.c_str());
          shortcut = new Shortcut(icon.c_str(), markup, style->GetHomeTileIconSize());
          shortcut->_id = TYPE_PLACE;
          if (filter != "")
          {
            std::vector<std::string> filter_key_value;
            boost::split(filter_key_value, filter, boost::is_any_of(":"));
            if (filter_key_value.size() == 2)
              shortcut->_place_id = g_strdup_printf("%s?filter_%s=%s",
                                                    source.c_str(),
                                                    filter_key_value[0].c_str(),
                                                    filter_key_value[1].c_str());
            else
              shortcut->_place_id = g_strdup(source.c_str());
          }
          else
            shortcut->_place_id = g_strdup(source.c_str());
          if (section != -1)
            shortcut->_place_section = section;
          _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
          shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
          g_free(markup);
        }
      }
    }
  }

  SetExpanded(true);
  SetCounts(8, 8);

  QueueDraw();
  _layout->QueueDraw();
  QueueRelayout();
}

void
PlacesHomeView::CreateShortcutFromExec(const char* exec,
                                       const char* name,
                                       std::vector<std::string>& alternatives)
{
  PlacesStyle*     style = PlacesStyle::GetDefault();
  Shortcut*        shortcut = NULL;
  gchar*           id;
  gchar*           markup;
  gchar*           icon;
  gchar*           real_exec;
  GDesktopAppInfo* info;

  markup = g_strdup_printf("<big>%s</big>", name);

  // We're going to try and create a desktop id from a exec string. Now, this is hairy at the
  // best of times but the following is the closest best-guess without having to do D-Bus
  // roundtrips to BAMF.
  if (exec)
  {
    char* basename;

    if (exec[0] == '/')
      basename = g_path_get_basename(exec);
    else
      basename = g_strdup(exec);

    id = g_strdup_printf("%s.desktop", basename);

    g_free(basename);
  }
  else
  {
    id = g_strdup_printf("%s.desktop", alternatives[0].c_str());
  }

  info = g_desktop_app_info_new(id);
  std::vector<std::string>::iterator iter = alternatives.begin();
  while (iter != alternatives.end())
  {
    if (!G_IS_DESKTOP_APP_INFO(info))
    {
      id = g_strdup_printf("%s.desktop", (*iter).c_str());
      info = g_desktop_app_info_new(id);
      iter++;
    }

    if (G_IS_DESKTOP_APP_INFO(info))
    {
      icon = g_icon_to_string(g_app_info_get_icon(G_APP_INFO(info)));
      real_exec = g_strdup(g_app_info_get_executable(G_APP_INFO(info)));

      shortcut = new Shortcut(icon, markup, style->GetHomeTileIconSize());
      shortcut->_id = TYPE_EXEC;
      shortcut->_exec = real_exec;
      _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);
      shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));

      g_free(icon);

      break;
    }
  }

  g_free(id);
  g_free(markup);
}

void PlacesHomeView::CreateShortcutFromMime(const char* mime,
                                            const char* name,
                                            std::vector<std::string>& alternatives)
{
  PlacesStyle* style = PlacesStyle::GetDefault();
  GAppInfo* info = g_app_info_get_default_for_type(mime, FALSE);

  // If it was invalid check alternatives for backup
  if (!G_IS_DESKTOP_APP_INFO(info))
  {
    for (auto alt: alternatives)
    {
      std::string id = alt + ".desktop";
      info = G_APP_INFO(g_desktop_app_info_new(id.c_str()));
      
      if (G_IS_DESKTOP_APP_INFO(info))
        break;
    }
  }

  if (G_IS_DESKTOP_APP_INFO(info))
  {
    glib::String icon(g_icon_to_string(g_app_info_get_icon(G_APP_INFO(info))));
    glib::String markup(g_strdup_printf("<big>%s</big>", name));
    
    Shortcut*   shortcut = new Shortcut(icon.Value(), markup.Value(), style->GetHomeTileIconSize());
    shortcut->_id = TYPE_EXEC;
    shortcut->_exec = g_strdup (g_app_info_get_executable(G_APP_INFO(info)));;
    shortcut->sigClick.connect(sigc::mem_fun(this, &PlacesHomeView::OnShortcutClicked));
    _layout->AddView(shortcut, 1, nux::eLeft, nux::eFull);

    g_object_unref(info);
  }
}

void
PlacesHomeView::OnShortcutClicked(PlacesTile* tile)
{
  Shortcut* shortcut = static_cast<Shortcut*>(tile);
  int id = shortcut->_id;

  if (id == TYPE_PLACE)
  {
    ubus_server_send_message(ubus_server_get_default(),
                             UBUS_PLACE_ENTRY_ACTIVATE_REQUEST,
                             g_variant_new("(sus)",
                                           shortcut->_place_id,
                                           shortcut->_place_section,
                                           ""));
  }
  else if (id == TYPE_EXEC)
  {
    GError* error = NULL;

    if (!g_spawn_command_line_async(shortcut->_exec, &error))
    {
      g_warning("%s: Unable to launch %s: %s",
                G_STRFUNC,
                shortcut->_exec,
                error->message);
      g_error_free(error);
    }

    ubus_server_send_message(ubus_server_get_default(),
                             UBUS_PLACE_VIEW_CLOSE_REQUEST,
                             NULL);
  }
}

const gchar* PlacesHomeView::GetName()
{
  return "PlacesHomeView";
}

const gchar*
PlacesHomeView::GetChildsName()
{
  return "";
}

void PlacesHomeView::AddProperties(GVariantBuilder* builder)
{
  unity::variant::BuilderWrapper(builder).add(GetGeometry());
}

} // namespace unity
